--- The utils module
-- @author Oskar Eisemuth
-- @copyright 2017-2018
-- @module commonapi.utils
local utils = {}

local osIsWin = (os.getenv("OS") == "Windows_NT")
local mymodid = nil

local function getOwnModId()
	local mypath = debug.getinfo(2, "S").source:sub(2)
	local mypath = string.gsub(mypath, '\\', '/')
	mymodid = string.match(mypath, "([^/]+)/res/scripts/commonapi/utils%.lua$")
end

getOwnModId()

--- Creates a log routine with prefix
-- @param prefix prefix for log
-- @return log function
function utils.logPrep(prefix)
	local prefix = prefix..":"
	local logfn = function(...)
		print(prefix, ...)
	end
	return logfn
end

local _log = utils.logPrep("commonapi.utils")

function utils.logPrepSlow(name)
	local prefix = name.."."
	local logfn = function(...)
		local dinfo = debug.getinfo(2, "n")
		if (dinfo ~= nil) then
			print(prefix..dinfo.name..":",...)
		else
			print(...)
		end
	end
	return logfn
end

--- Is running on windows.
-- @return boolean
function utils.isWindowsOS()
	return osIsWin
end


--- Get a valid tempfile.
-- Returns a valid tempfile name (on windows including temp directory)
-- @return string
function utils.getTempFilename() 
	local filename
	if (osIsWin == true) then
		local prefix = os.getenv("TEMP")
		filename = prefix..os.tmpname()
	else 
		filename = os.tmpname()
	end
	return filename
end


--- Returns the data function result of a game file.
-- @param filename A complete filepath to a lua file
-- @param env	Environment table (optional) 
-- @return table
function utils.getFileData(filename, env)
	return utils._getFileDataInternal(filename, env, true)
end

function utils._getFileDataIfExist(filename, env)
	return utils._getFileDataInternal(filename, env, false)
end

function utils._getFileDataInternal(filename, env, errIfMissing)
	local result = nil
	local luafn, err
	local ok

--	print("utils._getFileDataInternal", filename, env)

	if (env == nil) then
		env = {
			data = function()
				return {}
			end
		}
	end

	local debugGetInfo = debug.getinfo
	debug.getinfo = function(...)
		local n = select('#', ...);
		local info = nil
		if (n < 3) then
			local arg1 = select(1, ...);
			local arg2 = select(2, ...);
			if (type(arg1) == "number") then
				arg1 = arg1 + 1
			end

			if (type(arg2) == "string") then
				arg2 = arg2.."S"
			else
				arg2 = "S"
			end
			info = debugGetInfo(arg1, arg2)
		else
			local arg1 = select('1', ...);
			local arg2 = select('2', ...);
			local arg3 = select('3', ...);
			if (type(arg2) == "number") then
				arg2 = arg2 + 2
			end

			if (type(arg3) == "string") then
				arg3 = arg3.."S"
			else
				arg3 = "S"
			end
			info = debugGetInfo(arg1, arg2, arg3)
		end
		if (info ~= nil) then
			if (info["short_src"] == "[C]") then
				return nil
			end
		end
		return info
	end

	setmetatable(env, {__index=_G})

	luafn, err = loadfile(filename, "bt", env)
	if (luafn == nil) then
		if (errIfMissing) then
			_log(err)
		end
		debug.getinfo = debugGetInfo
		return nil
	end

	ok, err = pcall(luafn)
	if (ok == false) then
		_log(err)
		debug.getinfo = debugGetInfo
		return nil
	end

	if (env.data ~= nil) then
		ok, result = pcall(env.data)
		if (ok == false) then
			_log(result)
			debug.getinfo = debugGetInfo
			return nil
		end
	end

	debug.getinfo = debugGetInfo
	return result
end

function utils.quoteParam(filename) 
	return '"'..string.gsub(filename, '"', '\\"')..'"'
end

local function scanDirNative(nativefs, directory, result)
--	_log("scanDirNative:",  directory)
	local realpath = nativefs.realpath(directory)
	if (realpath ~= nil) then
		realpath = string.gsub(realpath, '\\', '/')..'/'
		for filename, fflags in nativefs.dir(directory) do
			if (fflags == 2) then
				scanDirNative(nativefs, directory..filename, result)
			else
				result[#result+1] = realpath..filename
			end
		end
	end
end

local function scanDirsNative(dirs)
	local result = {}
	local nativefs = commonapi._native.fs
	for id, entry in ipairs( dirs ) do
		scanDirNative(nativefs, entry, result)
	end
	return result
end

local function scanDirsShell(dirs)
	local result = {}
	local p
	
	if (dirs == nil) then
		return result
	end
	
	local ext
	
	if (osIsWin == true) then
		ext = ".cmd"
	else
		ext = ".sh"
	end
	
	local searchpath = string.gsub(package.path, "res/scripts/%?%.lua", "res/scripts/commonapi/?"..ext)
	
	local exefilename, err = package.searchpath("listdirs", searchpath)
	if (exefilename == nil) then
		_log(err)
		return result
	end
	
	_log(exefilename)
	if (string.match(exefilename, '^%a%a+') ~= nil) then
		_log("Adding local prefix to path")
		exefilename = "./"..exefilename
	end
	
	local tmpfilename = utils.getTempFilename()
	local f = assert(io.open(tmpfilename, "w"))
	for id, entry in ipairs( dirs ) do
		f:write(entry, "\n")
	end
	f:close()
	
	--local pname = '"'..exefilename..'" "'..tmpfilename..'"'
	local pcmd

	if (osIsWin == true) then
		pcmd = '"'..utils.quoteParam(exefilename)..' '..utils.quoteParam(tmpfilename)..'"'
	else
		pcmd = 'sh '..utils.quoteParam(exefilename)..' '..utils.quoteParam(tmpfilename)
	end
	
	_log("Trying to run "..pcmd)
	p = assert(io.popen(pcmd))
	for line in p:lines() do
		result[#result+1] = string.gsub(line, '\\', '/')
	end
	p:close()
	return result
end


--- Scan directories.
-- @param dirs A complete list of filepaths
-- @treturn {string,...} result of directory scan
function utils.scanDirs(dirs)
	local result = {}
	local p
	
	
	local testtimer = os.clock()
	if (dirs == nil) then
		return result
	end
	if (commonapi._native ~= nil) then
		result = scanDirsNative(dirs)
	else
		result = scanDirsShell(dirs)
	end

	_log(string.format("elapsed time for scanDirs: %.2f  (files %d)", os.clock() - testtimer, #result))
	return result
end

--- Create a ordered table
-- @return {table} a table with ordered key value iteration
function utils.createOrderedTable()
	local t = {}
	local mt = {
		__index = {
			_keyorder = {},
			_delete = function(self, key)
				if self[key] == nil then
					return
				end
				self[key] = nil
				local keyorder = self._keyorder
				for i, k in ipairs(keyorder) do
					if k == key then
						table.remove(keyorder, i)
					end
				end
			end,
			_getKeyIndex = function(self, key)
				local keyorder = self._keyorder
				for i, k in ipairs(keyorder) do
					if k == key then
						return i
					end
				end
				return nil
			end,
		},
		__newindex = function(self, key, value)
			rawset(self, key, value)
			if (value == nil) then
				return
			end
			local keyorder = self._keyorder
			keyorder[#keyorder+1] = key
		end,
		__pairs = function(tbl)
			local i = 0
			local function pairs_iter(self)
				i = i + 1
				local key = self._keyorder[i]
				if (key) then
					return key, self[key]
				end
				return
			end
			return pairs_iter, tbl, nil
		end
	}
	setmetatable(t, mt)
	return t
end

local function deepcopy(o)
	if ( type(o) == "table" ) then
		local newtable = {}
		for k, v in pairs(o) do
			if ( type(v) == "table" ) then
				newtable[k] = utils.deepcopy(v)
			else
				newtable[k] = v
			end
		end
		return newtable
	end
	return o
end

--- Copy a table
-- @return {table} a table deep copy
function utils.deepcopy(o)
	return deepcopy(o)
end

function utils.forceStrict()
	local mt = getmetatable(_G)
	if (mt == nil) then
		mt = {}
		setmetatable(_G, mt)
	end

	function mt.__newindex(t, key, value)
		local info = debug.getinfo(2, "S")
		if (info ~= nil) then
			if info.what == "main" or  info.what == "C" then
				rawset(t, key, value)
				return
			end
		end
		print("")
		print("STRICT: Attempt to write to global undeclared variable "..key)
		print(debug.traceback(nil, 2))
		rawset(t, key, value)
	end
	
	function mt.__index(_, key)
		print("")
		print("STRICT: Attempt to read global undeclared variable "..key)
		print(debug.traceback(nil, 2))
	end
end

function utils._getFunctionLocal(fnc, varname)
	local i = 1
	if fnc == nil then
		return nil
	end
	repeat
		local name, val = debug.getupvalue (fnc, i)
		if name == varname then
			return val
		end
		i = i + 1
	until not name
	return nil
end

function utils._dumpFunctionLocal(fnc, varname)
	local i = 1
	repeat
		local name, val = debug.getupvalue (fnc, i)
		print(name, val)
		i = i + 1
	until not name
	return nil
end



local detectedLanguage = nil
local _i18nLocal = {}
local _translateIdMap = {}

function utils._detectLanguage()
	local lang = string.match(debug.getinfo(_, "S").source, '%"([%w_]+)%", s%) end$')
	if (lang ~= nil) then
		if (detectedLanguage == lang) then
			_log("Detected language:", lang, "(same)")
		else
			_log("Detected language:", lang, "(new)")
			detectedLanguage = lang;
		end
	else
		_log("Detected language:", detectedLanguage, "(none=keepold)")
	end
end

function utils._addTranslateIdAlias(idalias, id)
	_translateIdMap[idalias] = id
end

-- Translate a string of modid
function utils.translateString(modid, s)
	local language
	if (detectedLanguage == nil) then
		language = string.match(debug.getinfo(_, "S").source, '%"([%w_]+)%", s%) end$') or "en"
	else
		language = detectedLanguage
	end
	
	if (_i18nLocal[modid] ~= nil) then
		return translate(modid, language, s)
	else
		if (_translateIdMap[modid] ~= nil) then
			return translate(_translateIdMap[modid], language, s)
		end
	end
	
--	print("utils.translateString", language, modid, s, "=", translate2(modid, language, s))
	return translate2(modid, language, s)
end

function utils._t(s)
	local language
	if (commonapi._native ~= nil) then
		language = commonapi._native.getLanguage()
	else
		language = string.match(debug.getinfo(_, "S").source, '%"([%w_]+)%", s%) end$') or "en"
	end
	return translate2(mymodid, language, s)
end

function prepareLanguageFunctions()
	if (setStrings == nil) then
		require "mod"
	end

	_i18nLocal = utils._getFunctionLocal(setStrings, "i18n")
	if (_i18nLocal == nil) then
		_log("ERROR: Can't get i18n local from setStrings!!!")
		_i18nLocal = {}
	end
end

prepareLanguageFunctions()

return utils
